@@ -10,11 +10,15 @@ from collections import namedtuple
 from operator import delitem
 import signal
 
+from zope.interface import implementer
+
+from gevent import getcurrent
+from gevent.exceptions import LoopExit
+
 from gevent._ffi import _dbg # pylint: disable=unused-import
 from gevent._ffi.loop import AbstractLoop
 from gevent._ffi.loop import assign_standard_callbacks
 from gevent._ffi.loop import AbstractCallbacks
-from gevent._util import implementer
 from gevent._interfaces import ILoop
 from gevent.libuv import _corecffi # pylint:disable=no-name-in-module,import-error
 
@@ -43,10 +47,6 @@ class _Callbacks(AbstractCallbacks):
 
         the_watcher.loop._queue_callback(watcher_ptr, revents)
 
-    def __loop_from_loop_ptr(self, loop_ptr):
-        loop_handle = loop_ptr.data
-        return self.from_handle(loop_handle)
-
 
 _callbacks = assign_standard_callbacks(
     ffi, libuv, _Callbacks,
@@ -112,7 +112,6 @@ class loop(AbstractLoop):
 
     def __init__(self, flags=None, default=None):
         AbstractLoop.__init__(self, ffi, libuv, _watchers, flags, default)
-        self.__loop_pid = os.getpid()
         self._child_watchers = defaultdict(list)
         self._io_watchers = dict()
         self._fork_watchers = set()
@@ -186,7 +185,17 @@ class loop(AbstractLoop):
                              self.SIGNAL_CHECK_INTERVAL_MS)
         libuv.uv_unref(self._signal_idle)
 
+    def __check_and_die(self):
+        if not self.ptr:
+            # We've been destroyed during the middle of self.run().
+            # This method is being called into from C, and it's not
+            # safe to go back to C (Windows in particular can abort
+            # the process with "GetQueuedCompletionStatusEx: (6) The
+            # handle is invalid.") So switch to the parent greenlet.
+            getcurrent().parent.throw(LoopExit('Destroyed during run'))
+
     def _run_callbacks(self):
+        self.__check_and_die()
         # Manually handle fork watchers.
         curpid = os.getpid()
         if curpid != self._pid:
@@ -320,6 +329,7 @@ class loop(AbstractLoop):
 
 
     def _stop_aux_watchers(self):
+        super(loop, self)._stop_aux_watchers()
         assert self._prepare
         assert self._check
         assert self._signal_idle
@@ -397,18 +407,21 @@ class loop(AbstractLoop):
             del self._fork_watchers
             del self._child_watchers
 
-
+    _HandleState = namedtuple("HandleState",
+                              ['handle',
+                               'type',
+                               'watcher',
+                               'ref',
+                               'active',
+                               'closing'])
     def debug(self):
         """
         Return all the handles that are open and their ref status.
         """
-        handle_state = namedtuple("HandleState",
-                                  ['handle',
-                                   'type',
-                                   'watcher',
-                                   'ref',
-                                   'active',
-                                   'closing'])
+        if not self.ptr:
+            return ["Loop has been destroyed"]
+
+        handle_state = self._HandleState
         handles = []
 
         # XXX: Convert this to a modern callback.
@@ -439,7 +452,8 @@ class loop(AbstractLoop):
         pass
 
     def break_(self, how=None):
-        libuv.uv_stop(self.ptr)
+        if self.ptr:
+            libuv.uv_stop(self.ptr)
 
     def reinit(self):
         # TODO: How to implement? We probably have to simply
@@ -483,19 +497,19 @@ class loop(AbstractLoop):
             val = _callbacks.python_callback(handle, arg)
             if val == -1: # Failure.
                 _callbacks.python_handle_error(handle, arg)
-            elif val == 1: # Success
+            elif val == 1: # Success, and we may need to close the Python watcher.
                 if not libuv.uv_is_active(watcher_ptr):
-                    # The callback closed the watcher in C. Good.
-                    # It's supposed to also reset the pointer to NULL at
+                    # The callback closed the native watcher resources. Good.
+                    # It's *supposed* to also reset the .data handle to NULL at
                     # that same time. If it resets it to something else, we're
                     # re-using the same watcher object, and that's not correct either.
-                    # Prevoiusly we checked for that case, but we shouldn't need to.
+                    # On Windows in particular, if the .data handle is changed because
+                    # the IO multiplexer is being restarted, trying to dereference the
+                    # *old* handle can crash with an FFI error.
                     handle_after_callback = watcher_ptr.data
                     try:
-                        if handle_after_callback:
+                        if handle_after_callback and handle_after_callback == handle:
                             _callbacks.python_stop(handle_after_callback)
-                        if handle_after_callback != handle:
-                            _callbacks.python_stop(handle)
                     finally:
                         watcher_ptr.data = ffi.NULL
         return True
@@ -560,6 +574,7 @@ class loop(AbstractLoop):
         return result
 
     def now(self):
+        self.__check_and_die()
         # libuv's now is expressed as an integer number of
         # milliseconds, so to get it compatible with time.time units
         # that this method is supposed to return, we have to divide by 1000.0
@@ -567,6 +582,7 @@ class loop(AbstractLoop):
         return now / 1000.0
 
     def update_now(self):
+        self.__check_and_die()
         libuv.uv_update_time(self.ptr)
 
     def fileno(self):
